<?php

namespace Drupal\workflow\Plugin\Field\FieldType;

use Drupal\Core\Entity\TypedData\EntityDataDefinition;
use Drupal\Core\Field\Attribute\FieldType;
use Drupal\Core\Field\EntityReferenceFieldItemList;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\OptGroup;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\ComplexDataDefinitionInterface;
use Drupal\Core\TypedData\DataDefinition;
use Drupal\Core\TypedData\DataReferenceDefinition;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\Core\Url;
use Drupal\options\Plugin\Field\FieldType\ListStringItem;
use Drupal\workflow\Entity\Workflow;
use Drupal\workflow\Entity\WorkflowInterface;
use Drupal\workflow\Entity\WorkflowState;
use Drupal\workflow\Entity\WorkflowTransition;
use Drupal\workflow\Entity\WorkflowTransitionInterface;
use Drupal\workflow\WorkflowTypeAttributeTrait;

/**
 * Defines the 'workflow' field, referencing a workflow_transition entity type.
 *
 * Supported settings (below the definition's 'settings' key) are:
 * - target_type: The entity type to reference. Required.
 */
#[FieldType(
  id: "workflow",
  label: new TranslatableMarkup("Workflow state"),
  description: [
    new TranslatableMarkup("Stores Workflow state ID's for a certain Workflow type from a list of allowed 'value => label' pairs"),
    new TranslatableMarkup("For example, 'Publishing': 1 => unpublished, 2 => draft, 3 => published"),
  ],
  category: "workflow",
  default_widget: "workflow_default",
  // default_widget: "entity_reference_autocomplete",
  default_formatter: "list_default",
  // default_formatter: "entity_reference_label",
  // list_class: EntityReferenceFieldItemList::class,
  cardinality: 1,
  constraints: [
    'WorkflowField' => [],
  ],
  module:"workflow",
)]

// class WorkflowItem extends EntityReferenceItem {
class WorkflowItem extends ListStringItem {

  use MessengerTrait;

  /*
   * Add variables and get/set methods for Workflow property.
   */
  use WorkflowTypeAttributeTrait;

  /**
   * {@inheritdoc}
   */
  public function __construct(ComplexDataDefinitionInterface $definition, $name = NULL, ?TypedDataInterface $parent = NULL) {
    parent::__construct($definition, $name, $parent);

    // Avoid exception in WorkflowTypeAttributeTrait::getWorkflowId();
    $this->setWorkflowId($this->getSetting('workflow_type'));
  }

  /**
   * {@inheritdoc}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition) {
    // $properties = $properties
    // + ListStringItem::propertyDefinitions($field_definition)
    // + EntityReferenceItem::propertyDefinitions($field_definition);
    $properties = parent::propertyDefinitions($field_definition);

    // Note: Settings are set in static::defaultStorageSettings();
    $settings = $field_definition->getSettings();

    // When Item is ListStringItem, ListItemBase.
      $properties['value'] = DataDefinition::create('string')
      ->setLabel(new TranslatableMarkup('Workflow state ID'))
      ->addConstraint('Length', ['max' => 128])
      ->setRequired(TRUE);

    // When Item is EntityReferenceItem.
    // Override parent integer ReferenceID for State key.
    $properties['target_id'] = DataDefinition::create('string')
      ->setLabel(t('Workflow state ID'))
      // @todo When Item is EntityReferenceItem, set to NOT computed (?).
      ->setComputed(TRUE)
      ->addConstraint('Length', ['max' => 128])
      ->setRequired(TRUE);
    $properties['entity'] = DataReferenceDefinition::create('entity')
      // = DataDefinition::create('WorkflowTransition')
      ->setLabel(t('State'))
      ->setDescription(new TranslatableMarkup('The Workflow state.'))
      // The entity object is computed out of the entity ID.
      ->setComputed(TRUE)
      ->setReadOnly(FALSE)
      ->setTargetDefinition(EntityDataDefinition::create($settings['target_type']));

      // Custom property for storing Transition, next to State.
      $properties['workflow_transition'] = DataReferenceDefinition::create('entity')
      ->setLabel(t('Transition'))
      ->setDescription(new TranslatableMarkup('The WorkflowTransition setting the Workflow state.'))
      // The entity object is computed out of the entity ID.
      ->setComputed(TRUE)
      ->setReadOnly(FALSE)
      ->setTargetDefinition(EntityDataDefinition::create($settings['target_transition']));

    return $properties;
  }

  /**
   * {@inheritdoc}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition) {
    // $schema = ListStringItem::schema($field_definition);
    // $schema = EntityReferenceItem::schema($field_definition);
    $schema = parent::schema($field_definition);

    $schema = [
      'columns' => [
        'value' => [
          'description' => 'The {workflow_states}.sid that this entity is currently in.',
          'type' => 'varchar',
          'length' => 128,
        ],
      ],
      'indexes' => [
        'value' => ['value'],
      ],
    ];
    return $schema;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultFieldSettings() {
    return [
      // 'handler' => 'default',
      // 'handler_settings' => [],
    ] + parent::defaultFieldSettings();
  }

  /**
   * {@inheritdoc}
   *
   * Note: Settings are used in static::propertyDefinitions();
   */
  public static function defaultStorageSettings() {
    $settings = [
      'workflow_type' => '',
      // When Item is ListStringItem, ListItemBase.
      'allowed_values' => [],
      'allowed_values_function' => 'workflow_state_allowed_values',
      // When Item is EntityReferenceItem.
      'target_type' => 'workflow_state',
      'target_bundle' => '',
      'handler' => 'default',
      // Custom property for storing Transition, next to State.
      'target_transition' => 'workflow_transition',
    ] + parent::defaultStorageSettings();
    return $settings;
  }

  /**
   * {@inheritdoc}
   */
  public function storageSettingsForm(array &$form, FormStateInterface $form_state, $has_data) {
    $element = parent::storageSettingsForm($form, $form_state, $has_data);

    $this->validateStorageSettingsForm($form, $form_state, $has_data);
    // Create list of all Workflow types. Include an initial empty value.
    $workflow_options = workflow_allowed_workflow_names(FALSE);
    $field_storage = $this->getFieldDefinition()->getFieldStorageDefinition();
    $wid = $this->getWorkflowId();

    // Set required workflow_type on 'comment' field of CommentWithWorkflow.
    /** @var \Drupal\Core\Field\FieldStorageDefinitionInterface $field_storage */
    if (!$wid && $field_storage->getTargetEntityTypeId() == 'comment') {
      $field_name = $field_storage->get('field_name');
      $workflow_options = [];
      foreach (_workflow_info_fields(NULL, '', '', $field_name) as $key => $info) {
        if (($info->getName() == $field_name) && ($info->getTargetEntityTypeId() !== 'comment')) {
          $wid = $info->getSetting('workflow_type');
          $workflow = Workflow::load($wid);
          $workflow_options[$wid] = $workflow->label();
        }
      }
    }

    // Let the user choose between the available workflow types.
    $url = Url::fromRoute('entity.workflow_type.collection')->toString();
    $element['workflow_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Workflow type'),
      '#options' => $workflow_options,
      '#default_value' => $wid,
      '#required' => TRUE,
      '#disabled' => $has_data,
      '#description' => $this->t('Choose the Workflow type. Maintain workflows
         <a href=":url">here</a>.', [':url' => $url]),
    ];

    // Overwrite ListItemBase::storageSettingsForm().
    // First, remove 'allowed values' list, due to restructured form in D10.2.
    unset($element['allowed_values']);
    // @todo Set 'allowed_values_function' properly in storage,
    // so default parent code can be used.
    // Do not change. @see https://www.drupal.org/project/drupal/issues/2643308
    $allowed_values_function = $this->defaultStorageSettings()['allowed_values_function'];
    $element['allowed_values_function'] = [
      '#type' => 'item',
      '#title' => $this->t('Allowed values list'),
      '#markup' => $this->t('The value of this field is being determined by the %function function and may not be changed.', ['%function' => $allowed_values_function]),
      '#access' => !empty($allowed_values_function),
      '#value' => $allowed_values_function,
    ];

    return $element;
  }

  /**
   * Generate messages on ConfigFieldItemInterface::settingsForm().
   */
  protected function validateStorageSettingsForm(array &$form, FormStateInterface $form_state, $has_data) {
    // Validate each workflow, and generate a message if not complete.
    // Create list of all Workflow types. Include an initial empty value.
    $workflow_options = workflow_allowed_workflow_names(FALSE);
    // @todo D8: add this to WorkflowFieldConstraintValidator.
    // Set message, if no 'validated' workflows exist.
    if (count($workflow_options) == 1) {
      $this->messenger()->addWarning(
        $this->t('You must <a href=":create">create at least one workflow</a>
          before content can be assigned to a workflow.',
          [':create' => Url::fromRoute('entity.workflow_type.collection')->toString()]
        ));
    }

    // Validate via annotation WorkflowFieldConstraint.
    // Show a message for each error.
    $violation_list = $this->validate();
    /** @var \Symfony\Component\Validator\ConstraintViolation $violation */
    foreach ($violation_list->getIterator() as $violation) {
      switch ($violation->getPropertyPath()) {
        case 'fieldnameOnComment':
          // @todo CommentWithWorkflow & storageSettingsForm() constraints.
          // A 'comment' field name MUST be equal to content field name.
          // @todo Fix fields on a non-relevant entity_type.
          $this->messenger()->addError($violation->getMessage());
          $workflow_options = [];
          break;

        default:
          break;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getEntity() {
    $entity = parent::getEntity();
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty() {
     // @todo $is_empty = parent::isEmpty();
    $is_empty = empty($this->value);
    return $is_empty;
  }

  /**
   * {@inheritdoc}
   *
   * Set both the Transition property AND the to_sid value.
   */
  public function setValue($values, $notify = TRUE) {
    $transition = NULL;
    $entity = $this->getEntity();
    $field_name = $this->getFieldName();

    switch (TRUE) {
      case $values instanceof WorkflowTransitionInterface:
        /** @var \Drupal\workflow\Entity\WorkflowTransitionInterface $transition */
        $transition = $values;
        $sid = $transition->isScheduled()
          ? $transition->getFromSid()
          : $transition->getToSid();
        break;

      case is_array($values) && isset($values['value']):
        // Add provisions for ReferencedEntity.
        $sid = $values['value'];
        break;

      default:
        // Add provisions for ReferencedEntity.
        $sid = $values;
        break;
    }

    // $items may be empty on node with options widget, or upon initial setting.
    $items = $entity->{$field_name} ?? NULL;
    if ($items) {
      $transition ??= WorkflowTransition::create([
        'entity' => $entity,
        'field_name' => $field_name,
        'wid' => $this->getWorkflowId(),
        'to_sid' => $sid,
      ]);
    }

    $values = [
      'value' => $sid,
      'target_id' => $sid,
      'workflow_transition' => $transition,
    ];
    parent::setValue($values, $notify);
  }

  /**
   * Gets the item's Field name.
   *
   * @return string
   *   The Field name.
   */
  public function getFieldName() {
    $field_name = $this->getParent()->getName();
    return $field_name;
  }

  /**
   * Gets the item's WorkflowState (of the first item in ItemList).
   *
   * @return \Drupal\workflow\Entity\WorkflowState
   *   The Transition object.
   */
  public function getState() {
    $sid = $this->getStateId();
    $state = WorkflowState::load($sid)
      ?? WorkflowState::create([
        'id' => $sid,
        'wid' => $this->getWorkflowId(),
      ]);
    return $state;
  }

  /**
   * Gets the item's WorkflowState ID (of the first item in ItemList).
   *
   * @return string
   *   The Workflow State ID.
   */
  public function getStateId() {
    $sid = $this->value;
    return $sid;
  }

  /**
   * Gets the item's WorkflowTransition (of the first item in ItemList).
   *
   * @return \Drupal\workflow\Entity\WorkflowTransitionInterface
   *   The Transition object, or NULL of not set.
   */
  public function getTransition() {
    $transition = NULL;

    $target_type = $this->getSetting('target_transition');
    $property = $this->get($target_type);
    /** @var \Drupal\workflow\Entity\WorkflowTransitionInterface $transition */
    $transition = $property->getValue();
    // Create a transition, to pass to the form.
    $transition ??= WorkflowTransition::create([
      'entity' => $this->getEntity(),
      'field_name' => $this->getParent()->getName(),
      'wid' => $this->getWorkflowId(),
    ]);
    return $transition;
  }

  /**
   * {@inheritdoc}
   */
  protected function allowedValuesDescription() {
    return '';
  }

  /**
   * Generates a string representation of an array of 'allowed values'.
   *
   * This string format is suitable for edition in a textarea.
   *
   * @param \Drupal\Core\Entity\EntityInterface[] $states
   *   An array of WorkflowStates, where array keys are values and array values are
   *   labels.
   * @param string $wid
   *   The requested Workflow ID.
   *
   * @return string
   *   The string representation of the $states array:
   *    - Values are separated by a carriage return.
   *    - Each value is in the format "value|label" or "value".
   */
  protected function allowedValuesString($states) {
    $lines = [];

    $wid = $this->getWorkflowId();

    $previous_wid = -1;
    /** @var \Drupal\workflow\Entity\WorkflowState $state */
    foreach ($states as $key => $state) {
      // Only show enabled states.
      if ($state->isActive()) {
        // Show a Workflow name between Workflows, if more then 1 in the list.
        if ((!$wid) && ($previous_wid <> $state->getWorkflowId())) {
          $previous_wid = $state->getWorkflowId();
          $workflow_label = $state->getWorkflow()->label();
          $lines[] = "$workflow_label's states: ";
        }
        $label = $this->t('@label', ['@label' => $state->label()]);

        $lines[] = "   $key|$label";
      }
    }
    return implode("\n", $lines);
  }

  /**
   * Implementation of TypedDataInterface.
   *
   * @see folder \workflow\src\Plugin\Validation\Constraint
   */

  /**
   * Implementation of OptionsProviderInterface.
   *
   *   An array of settable options for the object that may be used in an
   *   Options widget, usually when new data should be entered. It may either be
   *   a flat array of option labels keyed by values, or a two-dimensional array
   *   of option groups (array of flat option arrays, keyed by option group
   *   label). Note that labels should NOT be sanitized.
   */

  /**
   * {@inheritdoc}
   */
  public function getPossibleValues(?AccountInterface $account = NULL) {
    // Flatten options first, because SettableOptions may have 'group' arrays.
    $flatten_options = OptGroup::flattenOptions($this->getPossibleOptions($account));
    return array_keys($flatten_options);
  }

  /**
   * {@inheritdoc}
   */
  public function getPossibleOptions(?AccountInterface $account = NULL) {
    $allowed_options = [];

    // When we are initially on the Storage settings form, no wid is set, yet.
    if ($workflow = $this->getWorkflow()) {
      $allowed_options = $workflow->getStates(WorkflowInterface::ALL_STATES);
    }

    return $allowed_options;
  }

  /**
   * {@inheritdoc}
   */
  public function getSettableValues(?AccountInterface $account = NULL) {
    // Flatten options first, because SettableOptions may have 'group' arrays.
    $flatten_options = OptGroup::flattenOptions($this->getSettableOptions($account));
    return array_keys($flatten_options);
  }

  /**
   * {@inheritdoc}
   */
  public function getSettableOptions(?AccountInterface $account = NULL) {
    $allowed_options = [];

    // When we are initially on the Storage settings form, no wid is set, yet.
    if (!$wid = $this->getWorkflowId()) {
      return $allowed_options;
    }

    // On Field settings page, no entity is set.
    if (!$entity = $this->getEntity()) {
      return $allowed_options;
    }

    // Get the allowed new states for the entity's current state.
    $transition = $this->getTransition();
    $allowed_options = $transition?->getSettableOptions($account, 'to_sid');
    // $field_name = $this->getFieldDefinition()->getName();
    // $state = $this->getState();
    // @done use workflow_state_allowed_values, for executed transition.
    // $allowed_options = $state?->getOptions($entity, $field_name, NULL, FALSE);

    return $allowed_options ?? [];
  }

}
